#!/usr/bin/perl
#
# Usage: useradd.pl -u UID -g GID -c COMMENT -d HOME -s SHELL NAME
#
# Copyright(c) 2004  IWAMURO Motonori
#

use strict;

use IO::File;

use vars qw($PID $ETC $PASSWD $SHADOW);

BEGIN {
    # $PID = "./useradd.pid";
    $PID = "/var/run/useradd.pid";

    # $ETC = ".";
    $ETC = "/etc";
    $PASSWD = "$ETC/passwd";
    $SHADOW = "$ETC/shadow";
}

END {
    unlink($PID) if (-f $PID);
}

sub lock_useradd {
    my $pid = undef;
    for (my $i = 0; $i < 10; $i++) {
	$pid = new IO::File($PID, O_CREAT|O_EXCL|O_WRONLY) and last;
	sleep(1);
    }
    unless ($pid) {
	$PID = undef;
	exit(1);
    }
    $pid->print($$, "\n");
    $pid->close;
}

sub read_entries {
    my $entries = {};
    local $_;
    my $pfh = new IO::File($PASSWD);
    while (<$pfh>) {
	my ($name, $uid) = (split(/:/), $_)[0, 2];
	$entries->{$name} = {uid => $uid, passwd => $_};
    }
    $pfh->close;
    $entries->{nobody}{uid} = 99 if (exists($entries->{nobody}));
    my $sfh = new IO::File($SHADOW);
    while (<$sfh>) {
	my $name = (split(/:/), $_)[0];
	$entries->{$name}{shadow} = $_;
    }
    $sfh->close;
    return $entries;
}

sub write_entries {
    my $entries = shift;
    system("/bin/cp", "-a", $PASSWD, "$PASSWD-");
    system("/bin/cp", "-a", $SHADOW, "$SHADOW-");
    umask(022);
    my $pfh = new IO::File("$PASSWD.$$", "w") or die "Can't open $PASSWD.$$ ($!)";
    umask(077);
    my $sfh = new IO::File("$SHADOW.$$", "w") or die "Can't open $SHADOW.$$ ($!)";
    foreach my $entry (sort {$a->{uid} <=> $b->{uid}} values %$entries) {
	$pfh->print($entry->{passwd});
	$sfh->print($entry->{shadow});
    }
    $sfh->close;
    $pfh->close;
    rename("$PASSWD.$$", $PASSWD);
    rename("$SHADOW.$$", $SHADOW);
}

sub update_entries {
    my ($name, $uid, $gid, $comment, $home, $shell) = @_;
    my $entries = read_entries;
    if (exists($entries->{$name})) {
	print STDERR "login name=[$name] has already exists.\nAbort.\n";
	exit(1);
    }
    my $passwd = join(":", $name, "x", $uid, $gid, $comment, $home, $shell)."\n";
    my $shadow = join(":", $name, "*", int(time / (24 * 60 * 60)), 0, 99999, 7, '', '', '')."\n";
    $entries->{$name} = {uid => $uid, passwd => $passwd, shadow => $shadow};
    write_entries($entries);
}

sub main {
    if (@_ == 0) {
	print STDERR "Usage: useradd.pl -u UID -g GID -c COMMENT -d HOME -s SHELL NAME\n";
	exit(1);
    }
    lock_useradd;
    my ($name, $uid, $gid, $comment, $home, $shell);
    while (@_) {
	my $arg = shift;
	if ($arg eq '-u') {
	    $uid = shift;
	} elsif ($arg eq '-g') {
	    $gid = shift;
	} elsif ($arg eq '-c') {
	    $comment = shift;
	} elsif ($arg eq '-d') {
	    $home = shift;
	} elsif ($arg eq '-s') {
	    $shell = shift;
	} elsif (!defined($name)) {
	    $name = $arg;
	} else {
	    unshift(@_, $arg);
	    print STDERR "Invalid arguments [@_]\nAbort.\n";
	    exit(1);
	}
    }
    $shell ||= '/bin/false';
    my @errors;
    if ($uid !~ /^\d+/) {
	push(@errors, "Invalid uid=[$uid].\n");
    }
    if ($gid !~ /^\d+/) {
	push(@errors, "Invalid gid=[$gid].\n");
    }
    if ($home eq '') {
	push(@errors, "Missing home directory.\n");
    }
    if ($comment =~ /[\x00-\x1F:\x7F-\xFF]/) {
	push(@errors, "Invalid comment=[$comment].\n");
    }
    if (! -x $shell) {
	push(@errors, "Invaild login shell=[$shell].\n");
    }
    if ($name !~ /^[-a-z][-a-z0-9]*$/) {
	push(@errors, "Invaild login name=[$name].\n");
    }
    if (@errors) {
	print STDERR @errors, "Abort.\n";
	exit(1);
    }
    update_entries($name, $uid, $gid, $comment, $home, $shell);
}

main(@ARGV) if ($0 !~ /\.t$/);

1;
